"""
Copyright 2022 Joaquín I. Aramendía <samsagax at gmail dot com>

    This file is part of hhfc.

    hhfc is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

    hhfc is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
with hhfc. If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import logging
from hhfc import config, fan, sensor, controller


def arg_parse():
    """Basic argument parsing"""
    parser = argparse.ArgumentParser(
        description="Hwmon Handheld Fan Controller (hhfc)"
    )
    parser.add_argument('-c', '--config-file',
                        action='store',
                        type=str,
                        default="fan_control.yaml",
                        help="Configuration file path"
                        )
    parser.add_argument('-m', '--monitor',
                        action='store_true',
                        help="Monitor mode. When set the fans duty cycles are not modified"
                        )
    parser.add_argument('-l', '--loglevel',
                        action='store',
                        type=str,
                        default="INFO",
                        help="Log level to output (DEBUG, INFO, WARNING, ERROR, CRITICAL)"
                        )

    return parser.parse_args()


def setup_logging(loglevel):
    """Setup logging level from command line parameters"""
    match loglevel:
        case 'DEBUG':
            level = logging.DEBUG
        case 'INFO':
            level = logging.INFO
        case 'WARNING':
            level = logging.WARNING
        case 'ERROR':
            level = logging.ERROR
        case 'CRITICAL':
            level = logging.CRITICAL
        case _:
            level = logging.INFO
    logging.basicConfig(level=level)


def main():
    """The main application logic"""
    args = arg_parse()

    # Setup logging
    setup_logging(args.loglevel)

    # Read configuration
    conf = config.Config(args.config_file)

    fan_list = []
    for fan_conf in conf.get_fans_config():
        fan_list.append(fan.Fan(fan_conf))

    sensor_list = []
    for sensor_conf in conf.get_sensors_config():
        sensor_list.append(sensor.Sensor(sensor_conf))

    interval = conf.get_interval()

    control = controller.Controller(fan_list, sensor_list, interval, args.monitor)
    control.run()


if __name__ == '__main__':
    main()



"""
Copyright 2022 Joaquín I. Aramendía <samsagax at gmail dot com>

    This file is part of hhfc.

    hhfc is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

    hhfc is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
with hhfc. If not, see <https://www.gnu.org/licenses/>.
"""

from . import util


class Sensor:
    """Class to represent a hwmon Sensor"""

    name: str
    driver_name: str
    temp_input: str
    divisor: float
    offset: float
    curve: dict

    def __init__(self, sensor_config: dict):
        full_path = util.find_driver_path(sensor_config["driver_name"])

        self.name = sensor_config["name"]
        self.driver_name = sensor_config["driver_name"]
        self.sensor_input = full_path + sensor_config["temp_input"]
        self.divisor = sensor_config["divisor"] if "divisor" in sensor_config else 1
        self.offset = sensor_config["offset"] if "offset" in sensor_config else 0

    def read_input(self) -> float:
        """Check if we are controlling this fan"""
        with open(self.sensor_input, "r", encoding="utf-8") as raw_reading:
            return float(raw_reading.read()) / self.divisor + self.offset

    def __str__(self) -> str:
        """String representation of the sensor"""
        return str(self.name) + ": " + str(self.read_input())



"""
Copyright 2022 Joaquín I. Aramendía <samsagax at gmail dot com>

    This file is part of hhfc.

    hhfc is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

    hhfc is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
with hhfc. If not, see <https://www.gnu.org/licenses/>.
"""

import yaml

# Default valoes for configuration
## Controller
DEFAULT_INTERVAL = 1.0

## Sensors
DEFAULT_SENSOR_DIVISOR = 1000
DEFAULT_SENSOR_OFFSET = 0

## Fans
DEFAULT_FAN_MIN_CONTROL_VALUE = 0
DEFAULT_FAN_MAX_CONTROL_VALUE = 255
DEFAULT_FAN_ALLOW_SHUTOFF = "no"
DEFAULT_FAN_MINIMUM_DUTY_CYCLE = 30


class Config:
    """Class to manage configuration files"""

    config: dict

    def __init__(self, file_path: str, auto_load: bool = False):
        self.file_path = file_path
        self.config = None
        if auto_load:
            self._read_configuration()

    def _read_configuration(self) -> None:
        """Users should not call this function directly, would be called when
        config is needed
        """
        with open(self.file_path, 'r', encoding='utf-8') as file:
            self.config = yaml.safe_load(file)

        # Set default values if not configured
        if "INTERVAL" not in self.config:
            self.config["INTERVAL"] = DEFAULT_INTERVAL

        for sensor in self.get_sensors_config():
            if "divisor" not in sensor:
                sensor["divisor"] = DEFAULT_SENSOR_DIVISOR
            if "offset" not in sensor:
                sensor["offset"] = DEFAULT_SENSOR_OFFSET

        for fan in self.get_fans_config():
            if "max_control_value" not in fan:
                fan["max_control_value"] = DEFAULT_FAN_MAX_CONTROL_VALUE
            if "min_control_value" not in fan:
                fan["min_control_value"] = DEFAULT_FAN_MIN_CONTROL_VALUE
            if "allow_shutoff" not in fan:
                fan["allow_shutoff"] = DEFAULT_FAN_ALLOW_SHUTOFF
            if "minimum_duty_cycle" not in fan:
                fan["minimum_duty_cycle"] = DEFAULT_FAN_MINIMUM_DUTY_CYCLE

    def get_full_config(self) -> dict:
        """Get entire read dictionary, used for debug, mostly"""
        if not self.config:
            self._read_configuration()

        return self.config

    def get_fans_config(self, fan_idx: int = None) -> dict:
        """Returns a dictionary with fan configuration data"""
        if not self.config:
            self._read_configuration()

        return self.config["FANS"][fan_idx] \
            if (fan_idx is not None) else \
            self.config["FANS"]

    def get_sensors_config(self, sensor_idx: int = None) -> dict:
        """Returns a dictionary with sensors configuration data"""
        if not self.config:
            self._read_configuration()

        return self.config["SENSORS"][sensor_idx] \
            if (sensor_idx is not None) else \
            self.config["SENSORS"]

    def get_interval(self) -> float:
        """Returns interval data from configuration"""
        if not self.config:
            self._read_configuration()

        return self.config["INTERVAL"]



"""
Copyright 2022 Joaquín I. Aramendía <samsagax at gmail dot com>

    This file is part of hhfc.

    hhfc is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

    hhfc is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
with hhfc. If not, see <https://www.gnu.org/licenses/>.
"""

import threading
import logging
from hhfc.fan import Fan
from hhfc.sensor import Sensor


class Controller:
    """Class to represent a Controller. It handles Fans according to Sensors"""

    fans: list[Fan]
    sensors: list[Sensor]
    loop_interval: float
    monitor: bool
    exit_loop: threading.Event

    def __init__(self, fans: list[Fan], sensors: list[Sensor], loop_interval: float, monitor=False):
        self.fans = fans
        self.sensors = sensors
        self.loop_interval = loop_interval
        self.monitor = monitor
        self.exit_loop = threading.Event()

    def _loop_iter(self) -> None:
        sensor_readings = {}
        for sensor in self.sensors:
            sensor_readings[sensor.name] = sensor.read_input()

        if self.monitor:
            logging.info("Sensor readings: %s", str(sensor_readings))
        else:
            logging.debug("Sensor readings: %s", str(sensor_readings))

        for fan in self.fans:
            fan_duty = []
            for sensor in fan.sensors:
                if sensor["name"] in sensor_readings:
                    name = sensor["name"]
                    fan_duty.append(fan.get_desired_duty_cycle(name, sensor_readings[name]))
                else:
                    logging.warning("Sensor '%s' has no value for fan '%s'",
                                    sensor['name'],
                                    fan.name
                                    )
            if not self.monitor:
                logging.debug("fan %s: setting duty cycle to %s", fan.name, int(max(fan_duty)))
                fan.set_duty_cycle(int(max(fan_duty)))
                logging.debug("fan %s: %s RPM", fan.name, fan.read_input())
            else:
                logging.info("fan %s: %s RPM (%i)",
                             fan.name,
                             fan.read_input(),
                             int(max(fan_duty))
                             )

    def _loop(self) -> None:
        """Main control loop"""
        while True:
            self._loop_iter()
            if self.exit_loop.wait(timeout=self.loop_interval):
                return

    def _take_over_fans(self):
        """Put all fans in manual mode"""
        for fan in self.fans:
            try:
                logging.info("Taking control of fan: %s", fan.name)
                fan.take_control()
            except Exception as exp:
                logging.error("Could not take control of fan: %s", fan.name)
                logging.error(exp)

    def _release_fans(self):
        """Put all fans in automatic mode"""
        for fan in self.fans:
            try:
                logging.info("Release control of fan: %s", fan.name)
                fan.release_control()
            except Exception as exp:
                logging.error("Could not release fan: %s", fan.name)
                logging.error(exp)

    def get_sensors_for_fan(self, fan_idx: int) -> None:
        """Get the list of sensors for given fan index"""
        return self.fans[fan_idx]["sensors"]

    def run(self):
        """Runs the controller thread. This does not exit until interrupted"""

        # Set up
        if self.monitor:
            logging.info("Monitor mode, not taking over fans")
        else:
            logging.info("Taking over fans")
            self._take_over_fans()

        self.exit_loop.clear()
        loop = threading.Thread(target=self._loop, daemon=False)

        logging.debug("Starting control loop")
        loop.start()

        try:
            loop.join()
        except (KeyboardInterrupt, SystemExit):
            logging.info("Got Interrupt signal, bye!")
            self.exit_loop.set()
        except IOError as ioerr:
            logging.error("Got IOError: %s", ioerr)
            self.exit_loop.set()
        except Exception as exp:
            logging.warning("Uncaught Exception: %s", exp)
            self.exit_loop.set()

        # Cleanup
        if self.monitor:
            logging.info("Monitor mode, not restoring fans to automatic mode")
        else:
            logging.info("Restoring fans to automatic mode")
            self._release_fans()



"""
Copyright 2022 Joaquín I. Aramendía <samsagax at gmail dot com>

    This file is part of hhfc.

    hhfc is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

    hhfc is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
with hhfc. If not, see <https://www.gnu.org/licenses/>.
"""

import os

def find_driver_path(driver_name: str) -> str:
    """Find hwmon driver with name `driver_name` and returns the full path to
    its directory
    """
    base_path = "/sys/class/hwmon/"
    for driver in os.listdir(base_path):
        full_path = base_path + driver + "/"
        with open(full_path + "name", encoding="utf-8") as driver_path:
            test_name = driver_path.read().strip()
        if driver_name == test_name:
            return full_path

    raise RuntimeError(f"Driver with name {driver_name} not found.")



"""
Copyright 2022 Joaquín I. Aramendía <samsagax at gmail dot com>

    This file is part of hhfc.

    hhfc is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

    hhfc is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
with hhfc. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from . import util

class Interpolator:
    """Interpolator of a curve between given points at initialization"""

    x_vals: list[int]
    y_vals: list[int]

    def __init__(self, x_vals: list[int], y_vals: list[int]):
        if not len(x_vals) == len(y_vals):
            raise ValueError("x_vals and y_vals need to be the same lenght")
        self.x_vals = x_vals
        self.y_vals = y_vals

    def _compute_l_poly_value(self, j: int, x: int) -> int:
        """Compute the j-th lagrange polynomial at point x"""
        k = len(self.x_vals)
        result = 1
        for m in range(k):
            if m == j:
                continue
            result *= (x - self.x_vals[m]) / (self.x_vals[j] - self.x_vals[m])
        return result

    def get_value(self, x: int) -> int:
        """Evaluate at x, if x is lower than the lowest x, return the lowest
        value
        """
        if x < self.x_vals[0]:
            return self.y_vals[0]
        if x > self.x_vals[-1]:
            return self.y_vals[-1]

        k = len(self.x_vals)
        result = 0
        for j in range(k):
            result += self.y_vals[j] * self._compute_l_poly_value(j, x)

        return result


class Fan:
    """Class to represent and control a fan"""

    name: str
    driver_name: str
    pwm_input: str
    pwm_enable: str
    fan_input: str
    min_val: int
    max_val: int
    allow_shutoff: bool
    min_allowed: int
    sensors: list[dict]
    interpolator: dict

    def __init__(self, fan_config: dict):
        full_path = util.find_driver_path(fan_config["driver_name"])

        self.name = fan_config["name"]
        self.driver_name = fan_config["driver_name"]
        self.fan_input = full_path + fan_config["fan_input"]
        self.pwm_enable = full_path + fan_config["handle"] + "_enable"
        self.pwm_input = full_path + fan_config["handle"]
        self.min_val = fan_config["min_control_value"]
        self.max_val = fan_config["max_control_value"]
        self.allow_shutoff = (fan_config["allow_shutoff"] == "yes")
        self.min_allowed = fan_config["minimum_duty_cycle"]
        self.sensors = fan_config["sensors"]
        self.interpolator = {
            sens["name"]:self._generate_interpolator(sens["name"]) for sens in self.sensors
        }

    def take_control(self) -> bool:
        """Atempt to take control of the fan from automatic control"""
        with open(self.pwm_enable, "r+", encoding="utf-8") as enable:
            enable.write("1")
        return self.check_control()

    def release_control(self) -> bool:
        """Atempt to take control of the fan from automatic control"""
        with open(self.pwm_enable, "r+", encoding="utf-8") as enable:
            enable.write("0")
        return not self.check_control()

    def check_control(self) -> bool:
        """Check if we are controlling this fan"""
        with open(self.pwm_enable, "r", encoding="utf-8") as enable:
            string = enable.read()
            return int(string) == 1

    def get_sensor_curve(self, sensor: str) -> dict:
        """Returns the curve for sensor."""
        for sens in self.sensors:
            if sens["name"] == sensor:
                return sens["curve"]
        raise ValueError(f"Not such sensor: {sensor}")

    def _generate_interpolator(self, sensor: str) -> Interpolator:
        """Returns a Interpolator object for given sensor name"""
        curve = self.get_sensor_curve(sensor)
        temps = [ column[0] for column in curve ]
        dutys = [ column[1] for column in curve ]
        return Interpolator(temps, dutys)

    def get_desired_duty_cycle(self, sensor: str, value: int) -> int:
        """Returns duty cycle for the current sensor state according to the
        specified curve
        """
        interp = self.interpolator[sensor].get_value(value)
        # Modify value with shutoff and min_value policy
        if interp <= self.min_allowed:
            if not self.allow_shutoff:
                interp = self.min_allowed
            else:
                interp = 0
        return interp

    def set_duty_cycle(self, duty_cycle: int) -> None:
        """Sets duty cycle for this fan in the range [min_value-max_value]
        for the hwmon interface of choice. The input value `duty_cycle` should
        be in the range [0-100]. This function will scale acordingly.
        If the fan has shut-off policy set to "no" then it won't output a lower
        value than "minimum_duty_cycle".
        If the fan has shut-off policy set to "yes" then any value below
        "minimum_duty_cycle" will make the fan to turn off completely.
        """
        if duty_cycle < 0 or duty_cycle > 100:
            raise ValueError("Duty cycle has to be in the range [0-100]")

        # Enforce shutoff and min_value policy
        if duty_cycle <= self.min_allowed:
            if not self.allow_shutoff:
                duty_cycle = self.min_allowed
            else:
                duty_cycle = 0

        # Scale value
        duty = duty_cycle * (self.max_val - self.min_val) / 100.0

        if not self.check_control():
            logging.warning("Cant write to the fan control file. Duty cycle setting may fail")

        with open(self.pwm_input, "w", encoding="utf-8") as pwm:
            pwm.write(str(int(duty)))

    def read_input(self) -> int:
        """Read input for this fan. The value units are not converted"""
        with open(self.fan_input, "r", encoding="utf-8") as inp:
            return int(inp.read())

    def __str__(self) -> str:
        return (self.name + ": " + str(self.read_input()))

def fan_from_config(fan_config: dict) -> Fan:
    """Generate a Fan object with the configration passed as dictionary"""
    needed_keys = set(["name",
                       "driver_name",
                       "handle",
                       "max_control_value",
                       "min_control_value",
                       "fan_input"
                       ])

    if not needed_keys.issubset(fan_config.keys()):
        raise ValueError(f"Missing configuration. Needed {needed_keys}, "
                         "have {list(fan_config[name].keys())}")

    return Fan(fan_config)
